@@ -205,8 +205,7 @@ def _mudraw(buffer, fmt) -> bytes:
 
         proc = run(
             ['mudraw', '-F', fmt, '-o', '-', tmp_in.name],
-            stdout=PIPE,
-            stderr=PIPE,
+            capture_output=True,
             check=True,
         )
         return proc.stdout
@@ -942,11 +941,11 @@ class Extend_ObjectMapping:
         except KeyError:
             return default
 
-    @augment_if_no_cpp
+    @augment_if_no_cpp  # Remove this shim when dropping pybind11 < 2.8.0 support
     def keys(self):
         return KeysView(self)
 
-    @augment_if_no_cpp
+    @augment_if_no_cpp  # Remove this shim when dropping pybind11 < 2.8.0 support
     def values(self):
         return (v for _k, v in self.items())
 
@@ -1055,6 +1054,10 @@ class Extend_Page:
             If *res* does not belong to the same `Pdf` that owns this page,
             a copy of *res* is automatically created and added instead. In previous
             versions, it was necessary to change for this case manually.
+
+        .. versionchanged:: 4.3.0
+            Returns the name of the overlay in the resources dictionary instead
+            of returning None.
         """
         if Name.Resources not in self.obj:
             self.obj.Resources = Dictionary()
@@ -1083,15 +1086,19 @@ class Extend_Page:
         return name
 
     def _over_underlay(
-        self, other, rect: Optional[Rectangle], under: bool = True
-    ) -> None:
+        self,
+        other,
+        rect: Optional[Rectangle],
+        under: bool,
+        push_stack: bool,
+        shrink: bool,
+        expand: bool,
+    ) -> Name:
         formx = None
         if isinstance(other, Page):
-            page = other
             formx = other.as_form_xobject()
         elif isinstance(other, Dictionary) and other.get(Name.Type) == Name.Page:
-            page = Page(other)
-            formx = page.as_form_xobject()
+            formx = Page(other).as_form_xobject()
         elif (
             isinstance(other, Stream)
             and other.get(Name.Type) == Name.XObject
@@ -1100,17 +1107,35 @@ class Extend_Page:
             formx = other
 
         if formx is None:
-            raise TypeError("other object is not something we can convert to FormX")
+            raise TypeError(
+                "other object is not something we can convert to Form XObject"
+            )
 
         if rect is None:
-            rect = Rectangle(page.trimbox)
+            rect = Rectangle(self.trimbox)
 
         formx_placed_name = self.add_resource(formx, Name.XObject)
-        cs = self.calc_form_xobject_placement(formx, formx_placed_name, rect)
+        cs = self.calc_form_xobject_placement(
+            formx, formx_placed_name, rect, allow_shrink=shrink, allow_expand=expand
+        )
+
+        if push_stack:
+            self.contents_add(b'q\n', prepend=True)  # prepend q
+            self.contents_add(b'Q\n', prepend=False)  # i.e. append Q
 
         self.contents_add(cs, prepend=under)
+        self.contents_coalesce()
+        return formx_placed_name
 
-    def add_overlay(self, other: Union[Object, Page], rect: Optional[Rectangle] = None):
+    def add_overlay(
+        self,
+        other: Union[Object, Page],
+        rect: Optional[Rectangle] = None,
+        *,
+        push_stack: bool = True,
+        shrink: bool = True,
+        expand: bool = True,
+    ) -> Name:
         """Overlay another object on this page.
 
         Overlays will be drawn after all previous content, potentially drawing on top
@@ -1120,29 +1145,86 @@ class Extend_Page:
             other: A Page or Form XObject to render as an overlay on top of this
                 page.
             rect: The PDF rectangle (in PDF units) in which to draw the overlay.
-                If omitted, this page's trimbox, cropbox or mediabox will be used.
+                If omitted, this page's trimbox, cropbox or mediabox (in that order)
+                will be used.
+            push_stack: If True (default), push the graphics stack of the existing
+                content stream to ensure that the overlay is rendered correctly.
+                Officially PDF limits the graphics stack depth to 32. Most
+                viewers will tolerate more, but excessive pushes may cause problems.
+                Multiple content streams may also be coalseced into a single content
+                stream where this parameter is True, since the PDF specification
+                permits PDF writers to coalesce streams as they see fit.
+            shrink: If True (default), allow the object to shrink to fit inside the
+                rectangle. The aspect ratio will be preserved.
+            expand: If True (default), allow the object to expand to fit inside the
+                rectangle. The aspect ratio will be preserved.
+
+        Returns:
+            The name of the Form XObject that contains the overlay.
 
         .. versionadded:: 2.14
+
+        .. versionchanged:: 4.0.0
+            Added the *push_stack* parameter. Previously, this method behaved
+            as if *push_stack* were False.
+
+        .. versionchanged:: 4.2.0
+            Added the *shrink* and *expand* parameters. Previously, this method
+            behaved as if ``shrink=True, expand=False``.
+
+        .. versionchanged:: 4.3.0
+            Returns the name of the overlay in the resources dictionary instead
+            of returning None.
         """
-        return self._over_underlay(other, rect, under=False)
+        return self._over_underlay(
+            other,
+            rect,
+            under=False,
+            push_stack=push_stack,
+            expand=expand,
+            shrink=shrink,
+        )
 
     def add_underlay(
-        self, other: Union[Object, Page], rect: Optional[Rectangle] = None
-    ):
+        self,
+        other: Union[Object, Page],
+        rect: Optional[Rectangle] = None,
+        *,
+        shrink: bool = True,
+        expand: bool = True,
+    ) -> Name:
         """Underlay another object beneath this page.
 
         Underlays will be drawn before all other content, so they may be overdrawn
         partially or completely.
 
+        There is no *push_stack* parameter for this function, since adding an
+        underlay can be done without manipulating the graphics stack.
+
         Args:
             other: A Page or Form XObject to render as an underlay underneath this
                 page.
             rect: The PDF rectangle (in PDF units) in which to draw the underlay.
-                If omitted, this page's MediaBox will be used.
+                If omitted, this page's trimbox, cropbox or mediabox (in that order)
+                will be used.
+            shrink: If True (default), allow the object to shrink to fit inside the
+                rectangle. The aspect ratio will be preserved.
+            expand: If True (default), allow the object to expand to fit inside the
+                rectangle. The aspect ratio will be preserved.
+
+        Returns:
+            The name of the Form XObject that contains the underlay.
 
         .. versionadded:: 2.14
+
+        .. versionchanged:: 4.2.0
+            Added the *shrink* and *expand* parameters. Previously, this method
+            behaved as if ``shrink=True, expand=False``. Fixed issue with wrong
+            page rect being selected.
         """
-        return self._over_underlay(other, rect, under=True)
+        return self._over_underlay(
+            other, rect, under=True, push_stack=False, expand=expand, shrink=shrink
+        )
 
     def contents_add(self, contents: Union[Stream, bytes], *, prepend: bool = False):
         """Append or prepend to an existing page's content stream.
@@ -1269,6 +1351,9 @@ class Extend_AttachedFileSpec:
         This function will automatically add a creation and modified date
         using the file system, and a MIME type inferred from the file's extension.
 
+        If the data required for the attach is in memory, use
+        :meth:`pikepdf.AttachedFileSpec` instead.
+
         Args:
             pdf: The Pdf to attach this file specification to.
             path: A file path for the file to attach to this Pdf.
@@ -1286,7 +1371,7 @@ class Extend_AttachedFileSpec:
             pdf,
             path.read_bytes(),
             description=description,
-            filename=str(path),
+            filename=str(path.name),
             mime_type=mime,
             creation_date=encode_pdf_date(
                 datetime.datetime.fromtimestamp(stat.st_ctime)
